<?php

namespace Drupal\Tests\securepages\Functional;

use Drupal\comment\Tests\CommentTestTrait;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Session\AnonymousUserSession;
use Drupal\Core\Url;
use Drupal\KernelTests\AssertContentTrait;
use Drupal\language\Entity\ConfigurableLanguage;
use Drupal\securepages\Securepages;
use Drupal\Tests\BrowserTestBase;

/**
 * Test Secure Pages redirects.
 *
 * @group securepages
 */
class SecurepagesTest extends BrowserTestBase {

  use AssertContentTrait;
  use CommentTestTrait;

  /**
   * {@inheritdoc}
   */
  protected static $modules = ['securepages', 'node', 'comment', 'locale', 'path'];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * List of default pages protected.
   *
   * @var array
   *   List of strings.
   */
  protected $pagesDefault = ['/node/add*', '/node/*/edit', '/node/*/delete', '/user', '/user/*', '/admin', '/admin/*'];

  /**
   * {@inheritdoc}
   */
  public function setUp(): void {
    parent::setUp();
    $this->drupalCreateContentType(['type' => 'article']);
    $this->addDefaultCommentField('node', 'article');
  }

  /**
   * Runs all the tests in a sequence to avoid multiple re-installs.
   */
  public function testSecurePages() {
    $this->_testSettingsForm();
    $this->_testMatch();
    $this->_testLocale();
    $this->_testAnonymous();
    $this->_testFormAlter();
    $this->_testCachedResponse();
    $this->_testPathAlias();

    // @todo given no GET q in Drupal 8 and otherwise the format tested not
    // being a valid URL, not sure how would securepages be used for open
    // redirection.
    // $this->_testOpenRedirect();
    $this->_testXHR();
    $this->_testRoles();
    $this->_testPathNorms();
  }

  /**
   * Test submitting the settings form.
   */
  public function _testSettingsForm() {
    $this->drupalLoginHttps($this->drupalCreateUser(['administer site configuration']));
    $this->drupalGet(Securepages::getUrl('securepages.admin_settings')->toString());
    $this->submitForm(
      ['enable' => 1],
      'Save configuration'
    );
    $this->assertSession()->responseContains(t('The configuration options have been saved.'));

    // Clean up.
    $this->drupalLogoutHttps();
  }

  /**
   * Tests the securepages_match() function.
   */
  public function _testMatch() {
    \Drupal::configFactory()->getEditable('securepages.settings')->set('ignore', ['*/autocomplete/*'])->save();

    $request = \Drupal::requestStack()->getCurrentRequest();
    $test_request_is_https = $request->isSecure();
    $this->assertEquals(Securepages::matchPath('/user'), !$test_request_is_https, 'user matches.');
    $this->assertEquals(Securepages::matchPath('/user/login'), !$test_request_is_https, 'user/login matches.');
    $this->assertEquals(Securepages::matchPath('/admin/modules'), !$test_request_is_https, 'path admin/modules matches.');
    $this->assertEquals(Securepages::matchPath('/node'), $test_request_is_https, 'path node does not match.');

    $request = \Drupal::requestStack()->getCurrentRequest();
    $this->assertEquals(Securepages::matchPath('/user/autocomplete/alice'), $test_request_is_https, 'autocomplete path is ignored.');

    // Clean up.
    \Drupal::configFactory()->getEditable('securepages.settings')->set('ignore', [])->save();
  }

  /**
   * Tests correct operation with locale module.
   */
  public function _testLocale() {
    $french = ConfigurableLanguage::createFromLangcode('fr');
    $french->save();
    $this->drupalGet('fr/user');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->addressEquals(Url::fromRoute('<front>', [], ['https' => TRUE, 'absolute' => TRUE])->toString() . 'fr/user/login');
    $this->assertTrue(strstr($this->url, 'fr/user'), t('URL contains language prefix.'));
  }

  /**
   * Tests for anonymous browsing with securepages.
   */
  public function _testAnonymous() {
    // Visit the home page (same as user page) with plain HTTP.
    $this->drupalGet('', ['https' => FALSE]);
    $this->assertSession()->statusCodeEquals(302);
    $this->assertSession()->addressEquals(Url::fromRoute('user.login', [], ['https' => TRUE, 'absolute' => TRUE]));

    // Visit the login page (same as front page in tests) with plain HTTP.
    $this->drupalGet('user/login', ['https' => FALSE]);
    $this->assertSession()->statusCodeEquals(302);
    $this->assertSession()->addressEquals(Url::fromRoute('user.login', [], ['https' => TRUE, 'absolute' => TRUE]));

    // Visit the home page and login with HTTPS.
    $this->drupalGet('', ['https' => TRUE]);
    $this->assertSession()->statusCodeEquals(302);
    $this->assertSession()->addressEquals(Url::fromRoute('user.login', [], ['https' => TRUE, 'absolute' => TRUE]));
    $this->drupalGet('user/login', ['https' => TRUE]);
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->addressEquals(Url::fromRoute('user.login', [], ['https' => TRUE, 'absolute' => TRUE]));

    // Enable "Switch back to http pages when there are no matches".
    \Drupal::configFactory()->getEditable('securepages.settings')->set('switch', TRUE)->save();

    // Visit the home page and user with HTTPS and confirm the switch-back.
    $this->drupalGet('', ['https' => TRUE]);
    $this->assertSession()->statusCodeEquals(302);
    $this->assertSession()->addressEquals(Url::fromRoute('user.login', [], ['https' => FALSE, 'absolute' => TRUE]));
    $this->drupalGet('user/login', ['https' => TRUE]);
    $this->assertSession()->statusCodeEquals(302);
    $this->assertSession()->addressEquals(Url::fromRoute('user.login', [], ['https' => FALSE, 'absolute' => TRUE]));

    // Clean up.
    \Drupal::configFactory()->getEditable('securepages.settings')->set('switch', FALSE)->save();
  }

  /**
   * Tests the ability to alter form actions.
   *
   * Uses the comment form, since it has an #action set.
   */
  public function _testFormAlter() {
    $config = \Drupal::configFactory()->getEditable('securepages.settings');
    $config->set('switch', TRUE)->save();

    // Enable anonymous user comments.
    user_role_change_permissions(AccountInterface::ANONYMOUS_ROLE, [
      'access comments' => TRUE,
      'post comments' => TRUE,
      'skip comment approval' => TRUE,
    ]);

    $account = $this->drupalCreateUser(['access content', 'access comments', 'post comments', 'skip comment approval']);
    $node = $this->drupalCreateNode(['type' => 'article', 'promote' => 1]);

    foreach (['anonymous', 'authenticated'] as $mode) {
      if ($mode == 'authenticated') {
        $this->drupalLogin($account);
      }

      // Test plain HTTP posting to HTTPS.
      $config->set('pages', ['/comment/reply/*', '/user*'])->save();
      $this->drupalGet('node/' . $node->id(), ['https' => FALSE]);
      $this->assertFieldByXPath('//form[@class="comment-form" and starts-with(@action, "https:")]', NULL, "The $mode comment form action is https.");
      $this->submitForm(['comment_body[0][value]' => 'test comment'], 'Save');
      $this->assertSession()->responseContains(t('Your comment has been posted.'));

      // Test HTTPS posting to plain HTTP.
      $config->set('pages', ['/node/*', '/user*'])->save();
      $this->drupalGet('node/' . $node->id(), ['https' => TRUE]);
      $this->assertSession()->addressEquals(Url::fromRoute('entity.node.canonical', ['node' => $node->id()], ['https' => TRUE, 'absolute' => TRUE]));
      $this->assertFieldByXPath('//form[@class="comment-form" and starts-with(@action, "http:")]', NULL, "The $mode comment form action is http.");
      $this->submitForm(['comment_body[0][value]' => 'test'], 'Save');
      $this->assertSession()->responseContains(t('Your comment has been posted.'));
    }
    $this->drupalLogout();

    // Test the user login block.
    $this->drupalGet('');
    $edit = [
      'name' => $account->getAccountName(),
      'pass' => $account->pass_raw,
    ];
    $this->submitForm($edit, 'Log in');
    $this->drupalGet('user/' . $account->id() . '/edit');
    $this->assertSession()->statusCodeEquals(200);

    // Clean up.
    $this->drupalLogout();
    $config
      ->set('switch', FALSE)
      ->set('pages', $this->pagesDefault)
      ->save();
  }

  /**
   *
   */
  public function _testCachedResponse() {
    // Enable the page cache and fetch the login page.
    $this->installModule('page_cache');
    $url = Url::fromRoute('user.login', [], ['absolute' => TRUE, 'https' => FALSE]);
    $this->drupalGet($url);

    // Short-circuit redirects within the simpletest browser.
    $maximum_redirects = $this->maximumRedirects;
    $this->maximumRedirects = 0;
    $this->drupalGet($url);
    $this->assertSession()->statusCodeEquals(302);
    $this->assertEquals($this->getSession()->getResponseHeader('Location'), Url::fromRoute('user.login', [], ['https' => TRUE, 'absolute' => TRUE])->toString());
    $this->assertEquals($this->getSession()->getResponseHeader('X-Drupal-Cache'), 'HIT', 'Page was cached.');

    // Clean up.
    $this->maximumRedirects = $maximum_redirects;
    $this->uninstallModule('page_cache');
  }

  /**
   * Test redirection on aliased paths.
   */
  public function _testPathAlias() {
    $config = \Drupal::configFactory()->getEditable('securepages.settings');
    $config->set('pages', ['/node/*', '/user*'])->save();

    // Create test user and login.
    $this->drupalLogin($this->drupalCreateUser(['administer url aliases', 'create url aliases']));

    // Create test node.
    $node = $this->drupalCreateNode();

    // Create alias.
    $edit = [];
    $edit['source'] = 'node/' . $node->id();
    $edit['alias'] = $this->randomMachineName();
    $this->drupalGet('admin/config/search/path/add');
    $this->submitForm($edit, 'Save');

    // Short-circuit redirects within the simpletest browser.
    $maximum_redirects = $this->maximumRedirects;
    $this->maximumRedirects = 0;
    $this->drupalGet($edit['alias'], ['absolute' => TRUE, 'https' => FALSE]);
    $this->assertSession()->statusCodeEquals(302);
    $this->assertEquals($this->getSession()->getResponseHeader('Location'), Url::fromRoute('<front>', [], ['https' => TRUE, 'absolute' => TRUE])->toString() . $edit['alias']);

    // Clean up.
    $this->maximumRedirects = $maximum_redirects;
    $this->drupalLogout();
    $config->set('pages', $this->pagesDefault)->save();
  }

  /**
   * Verifies that securepages is not an open redirect.
   */
  public function _testOpenRedirect() {
    // Short-circuit redirects within the simpletest browser.
    \Drupal::state()->set('simpletest_maximum_redirects', 0);
    \Drupal::state()->set('securepages_switch', TRUE);

    global $base_url, $base_path;
    $secure_base_url = str_replace('http', 'https', $base_url);
    $this->drupalGet($secure_base_url . $base_path . '?q=http://example.com/', ['external' => TRUE]);
    $this->assertSession()->statusCodeEquals(302);
    $this->assertTrue(strstr($this->getSession()->getResponseHeader('Location'), $base_url), t('Open redirect test passed.'));

    $this->drupalGet($secure_base_url . $base_path . '?q=' . urlencode('http://example.com/'), ['external' => TRUE]);
    $this->assertSession()->statusCodeEquals(302);
    $this->assertTrue(strstr($this->getSession()->getResponseHeader('Location'), $base_url), t('Open redirect test passed.'));

    // Clean up.
    \Drupal::state()->delete('simpletest_maximum_redirects');
    \Drupal::state()->delete('securepages_switch');
  }

  /**
   * Test detection of XHR requests.
   */
  public function _testXHR() {
    // Without XHR header.
    $this->drupalGet('user', ['https' => FALSE]);
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->addressEquals(Url::fromRoute('user.login', [], ['https' => TRUE, 'absolute' => TRUE]));

    // With XHR header.
    $this->drupalGet('user', ['https' => FALSE], ['X-Requested-With: XMLHttpRequest']);
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->addressEquals(Url::fromRoute('user.login', [], ['https' => FALSE, 'absolute' => TRUE]));
  }

  /**
   * Test role-based switching.
   */
  public function _testRoles() {
    $account = $this->drupalCreateUser(['access content']);
    $role = current($account->getRoles(TRUE));

    $this->drupalLoginHttps($account);
    $config = \Drupal::configFactory()->getEditable('securepages.settings');
    $config->set('switch', TRUE)->set('roles', [$role])->set('pages', ['/admin*'])->save();

    // Visit the home page and /user with HTTPS and confirm that redirection happens.
    $this->drupalGet('', ['https' => FALSE]);
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->addressEquals(Url::fromRoute('<front>', [], ['https' => TRUE, 'absolute' => TRUE]));
    $this->drupalGet('user', ['https' => FALSE]);
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->addressEquals(Url::fromRoute('user.page', [], ['https' => TRUE, 'absolute' => TRUE]));

    // Test that forms actions aren't switched back to http.
    $node = $this->drupalCreateNode(['type' => 'article', 'promote' => 1]);
    $this->drupalGet('node/' . $node->id(), ['https' => TRUE]);
    $this->assertFieldByXPath('//form[@class="comment-form" and starts-with(@action, "/")]', NULL, "The comment form action is https.");

    // Clean up.
    $config->set('switch', FALSE)->set('roles', [])->set('pages', $this->pagesDefault)->save();
    $this->drupalLogout();
  }

  /**
   * Test path normalization checks.
   */
  public function _testPathNorms() {
    $config = \Drupal::configFactory()->getEditable('securepages.settings');
    $config->set('switch', TRUE)->set('pages', ['/user'])->save();

    // Test mixed-case path.
    $this->drupalGet('UsEr');
    $this->assertSession()->addressEquals('UsEr', ['https' => TRUE, 'absolute' => TRUE]);
    $this->assertFieldByXPath('//form[@id="user-login" and starts-with(@action, "/")]', NULL, 'The user login form action is https.');

    // Test that a trailing slash will not force a protected form's action to
    // http. A http based 'user/' path will become 'user' when doing the
    // redirect, so best to ensure that the test gets the right conditions the
    // path should be https based.
    $this->drupalGet('user/', ['https' => TRUE, 'absolute' => TRUE]);
    $this->assertSession()->addressEquals('user/', ['https' => TRUE, 'absolute' => TRUE]);
    $this->assertFieldByXPath('//form[@id="user-login" and starts-with(@action, "/")]', NULL, 'The user login form action is https.');

    // Clean up.
    $config->set('switch', FALSE)->set('pages', $this->pagesDefault)->save();
  }

  /**
   * Log in a user with the internal browser using HTTPS.
   *
   * @see \Drupal\simpletest\WebTestBase::drupalLogin()
   *
   * @param \Drupal\Core\Session\AccountInterface $account
   *   User object representing the user to log in.
   */
  public function drupalLoginHttps($account) {
    if ($this->loggedInUser) {
      $this->drupalLogout();
    }

    $edit = [
      'name' => $account->getAccountName(),
      'pass' => $account->pass_raw,
    ];
    $this->drupalGet(Securepages::getUrl('user.login')->toString());
    $this->submitForm($edit, 'Log in');

    // @see WebTestBase::drupalUserIsLoggedIn()
    if (isset($this->sessionId)) {
      $account->session_id = $this->sessionId;
    }
    $pass = $this->assertSession()->responseContains('>' . $account->getAccountName() . '<', t('User %name successfully logged in.', ['%name' => $account->getAccountName()]), 'User login');
    if ($pass) {
      $this->loggedInUser = $account;
      $this->container->get('current_user')->setAccount($account);
    }
  }

  /**
   * Logs a user out of the internal browser and confirms.
   *
   * Confirms logout by checking the login page.
   */
  protected function drupalLogoutHttps() {
    // Make a request to the logout page, and redirect to the user page, the
    // idea being if you were properly logged out you should be seeing a login
    // screen.
    $this->drupalGet(Securepages::getUrl('user.logout', ['query' => ['destination' => 'user/login']])->toString());
    $this->assertSession()->statusCodeEquals(200);
    $pass = $this->assertField('name', 'Username field found.', 'Logout');
    $pass = $pass && $this->assertField('pass', 'Password field found.', 'Logout');

    if ($pass) {
      // @see WebTestBase::drupalUserIsLoggedIn()
      unset($this->loggedInUser->session_id);
      $this->loggedInUser = FALSE;
      $this->container->get('current_user')->setAccount(new AnonymousUserSession());
    }
  }

  /**
   * Installs a module.
   *
   * @param string $module
   *   The module name.
   */
  protected function installModule($module) {
    $this->container->get('module_installer')->install([$module]);
    $this->container = \Drupal::getContainer();
  }

  /**
   * Uninstalls a module.
   *
   * @param string $module
   *   The module name.
   */
  protected function uninstallModule($module) {
    $this->container->get('module_installer')->uninstall([$module]);
    $this->container = \Drupal::getContainer();
  }

}
